#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
#
#
#
# Dependencies:  gspread, oauth2client, caldav

import gspread
# import json
import os
import time
from calendar import TimeEncoding, month_name
from datetime import date, datetime, timedelta
from oauth2client.service_account import ServiceAccountCredentials
from ics import Calendar, Event
from ics.parse import ContentLine
import ConfigParser
import SyncToCalDav
import re

# fix for utf-8 encoding
import sys
reload(sys)
sys.setdefaultencoding('utf8')
ENCODING = 'de_DE.UTF-8'

GOOGLE_CREDENTIALS_FILE = os.path.join(os.getenv('HOME'), '.google.json')
GOOGLE_KEY = '1i0DOGO9AXg8kKq0W7xFLqwgse1tav1wgxtKkx6XCCqA' #NEW
#GOOGLE_KEY = '11r_98CcYxf8zN7-D3KejM2mxhUyhft4TvJ9RAJlqnz4' #OLD
SCOPE = ['https://spreadsheets.google.com/feeds']


CALDAV_CREDENTIALS_FILE = os.path.join(os.getenv('HOME'), '.calendar.secret')
CALDAV_CALENDAR_ID = '1349'


def check_needed_files():
	if not os.path.isfile(GOOGLE_CREDENTIALS_FILE):
		raise Exception("%s not found, exiting." % (GOOGLE_CREDENTIALS_FILE,))
	if not os.path.isfile(CALDAV_CREDENTIALS_FILE):
		raise Exception("%s not found, exiting." % (CALDAV_CREDENTIALS_FILE,))


def sanitize_ints(func):
	def decorate(self, *args, **kwargs):
		try:
			numbers = [int(n) for n in args]
		except:
			raise(ValueError("Can't convert to int"))
		else:
			return func(self, *tuple(numbers), **kwargs)
	return decorate


def sanitize_strings(func):
	def decorate(self, *args, **kwargs):
		try:
			strings = [str(s) for s in args]
		except:
			raise(ValueError("Can't convert to string"))
		else:
			return func(self, *tuple(strings), **kwargs)
	return decorate


def sanitize_dates(func):
	def decorate(*args, **kwargs):
		for possible_date in args:
			if not isinstance(possible_date, datetime):
				raise (ValueError("Not a date"))
		return func(*args, **kwargs)
	return decorate


class Cal:

	def __init__(self):
		self.current_year = time.localtime().tm_year
		self.current_month = time.localtime().tm_mon

	@ sanitize_ints
	def day_count(self, year, month):
		_month = self.get_month_start_end(year, month)
		return _month.get('end')

	@sanitize_ints
	def get_month_start_end(self, year, month):
		date_range = {}
		date_range['start'] = date(year, month, 1)
		date_range['end'] = self.get_last_day_of_month(year, month)
		return date_range

	@sanitize_ints
	def get_month_lenght(self, year, month):
		last = self.get_last_day_of_month(year, month)
		result = int(last.strftime("%d"))
		return result

	@sanitize_ints
	def get_last_day_of_month(self, year, month, ret_format=0):
		next_month_year, next_month_month = self.plus_one_month(year, month)
		_tmp_date = date(next_month_year, next_month_month, 1)
		end_date = _tmp_date - timedelta(days=1)
		if (ret_format == 0):
			return end_date
		else:
			return end_date.day

	@sanitize_dates
	def date_range(self, start_date, end_date):
		start_date = start_date.date()
		end_date = end_date.date()
		for n in range(self.day_count if self.day_count else 1):
			yield start_date + timedelta(n)

	def divide_month_into_weeks(self, month_range):
		month_start = month_range['start']
		month_end = month_range['end']
		weeks = {}
		for day in self.date_range(month_start, month_end):
			week_nr = str(int(day.strftime("%W")) + 1)
			if week_nr not in weeks:
				weeks[week_nr] = {}
				weeks[week_nr]['days'] = []
				weeks[week_nr]['holidays'] = []
			if day in self.holidays:
				weeks[week_nr]['holidays'].append(day)
			else:
				weeks[week_nr]['days'].append(day)
		return weeks

	def get_month_name(self, month_no, locale=ENCODING):
		if isinstance(month_no, str):
			month_no = int(month_no)
		with TimeEncoding(locale) as encoding:
			name = month_name[month_no]
			if encoding is not None:
				name = name.decode(encoding)
			return name

	@sanitize_ints
	def plus_one_day(self, year, month, day):
		start_date = date(year, month, day)
		next_day = start_date + timedelta(days=1)
		return next_day.year, next_day.month, next_day.day

	@sanitize_ints
	def plus_one_month(self, year, month):
		start_date = date(year, month, 1)
		next_month = start_date + timedelta(days=33)
		return next_month.year, next_month.month

	@sanitize_ints
	def minus_one_month(self, year, month):
		start_date = date(year, month, 1)
		previous_month = start_date - timedelta(days=1)
		return previous_month.year, previous_month.month

	@sanitize_ints
	def minus_six_months(self, year, month):
		_year = year
		_month = month
		for i in range(1, 7):
			_year, _month = self.minus_one_month(_year, _month)
		return _year, _month

	@sanitize_ints
	def plus_six_months(self, year, month):
		_year = year
		_month = month
		for i in range(1, 7):
			_year, _month = self.plus_one_month(_year, _month)
		return _year, _month

	def get_calendar_start_end_dates(self):
		cal_min_year, cal_min_month = self.minus_one_month(self.current_year, self.current_month)
		#cal_max_year, cal_max_month = self.plus_six_months(self.current_year, self.current_month)
		cal_max_year, cal_max_month = self.current_year+1, 12
		return {'year_min': cal_min_year, 'month_min': cal_min_month, 'year_max': cal_max_year, 'month_max': cal_max_month}

	def month_range_named(self, start, end=12):
		_month_range = range(start, end + 1)
		return [self.get_month_name(month) for month in _month_range]

	def month_range(self, start, end=12):
		return range(start, end + 1)

	def get_calendar_month_range(self):
		_range = self.get_calendar_start_end_dates()
		if _range.get('year_min') == _range.get('year_max'):
			month_range = self.month_range(_range.get('month_min'), _range.get('month_max'))
			result = {_range.get('year_min'): month_range}
		else:
			result = {}
			result[_range.get('year_min')] = self.month_range(_range.get('month_min'))
			result[_range.get('year_max')] = self.month_range(1, _range.get('month_max'))
		return result


class Spreadsheet:

	def __init__(self):
		self.credentials = ServiceAccountCredentials.from_json_keyfile_name(GOOGLE_CREDENTIALS_FILE, SCOPE)
		self.gc = gspread.authorize(self.credentials)
		self.sheet = self.gc.open_by_key(GOOGLE_KEY)

	def gather_data(self):
		data = self.get_months_data(self.named_month)
		return data

	@sanitize_strings
	def get_worksheet(self, year):
		'''
		instanciate worksheet for given year
		'''
		sheet = None
		try:
			sheet = self.sheet.worksheet(year)
		except Exception as e:
			#print(e)
			#raise Exception("Could not get Spreadsheet Data")
			return None
		return sheet

	def get_months_data(self, worksheet, year, month):
		'''
		gather spreadsheet data for one month
		'''
		named_month = cal.get_month_name(month)
		base_cell = worksheet.find(named_month)
		data = {}
		day_last = cal.get_last_day_of_month(year, month, 1)
		cells = worksheet.range(base_cell.row + 1, base_cell.col, base_cell.row + 1 + day_last, base_cell.col)
		for day in range(1, day_last+1):
			data[day] = cells[day-1].value
		return data

	def get_data_for_months(self, year, months):
		'''
		gathers data of month range for specified year from spreadsheet, returns data in
		"{'<month>'{<'day'>: <'name'>, ...}, ...}" format
		'''
		data = {}
		worksheet = self.get_worksheet(year)
		if (worksheet == None):
			return None
		for month in months:
			data[month] = self.get_months_data(worksheet, year, month)
		return data


class Ical:

	def __init__(self, data, cal, config):

		self.data = data
		self.filename = '24_7.calendar'
		self.cal = cal
		self.calendar = Calendar()
		self.dav = None
		self.event_list = []
		self.init_dav(config)
		self.range_start = None
		self.range_end = None

	def init_dav(self, config):
		url = config.get('CALDAV', 'URL')
		password = config.get('CALDAV', 'PASSWORD')
		user = config.get('CALDAV', 'USER')
		self.dav = SyncToCalDav.SyncToCalDav(url=url, password=password, username=user, calendar_id=CALDAV_CALENDAR_ID)

	def set_data_range(self):
		for year in self.data.iterkeys():
			for month in self.data[year].iterkeys():
				for day in self.data[year][month].iterkeys():
					if(self.data[year][month][day]):
						if(not self.range_start):
							self.range_start = {"year": year, "month": month, "day": day}
						elif(self.range_end):
							self.range_end = None
					elif(self.range_start and not self.range_end):
						self.range_end = {"year": year, "month": month, "day": day-1}
	def build_cal_structure(self):
		'''
		parse spreadsheet data and build events and append them to self.calendar
		'''
		self.set_data_range()
		
		# iterate over all days and try to concat them to create less
		# single appointments
		_start_day = None
		_end_day = None

		for year in self.data.iterkeys():
			if(year < self.range_start["year"]):
				continue
			for month in self.data[year].iterkeys():
				if(year == self.range_start["year"] and month < self.range_start["month"]):
					continue
				days = self.data[year][month].keys()
				for day in days:
					if(year == self.range_start["year"] and month == self.range_start["month"] and day < self.range_start["day"]):
						continue
					elif(year == self.range_end["year"] and month == self.range_end["month"] and day > self.range_end["day"]):
						_end_day = _day
						_date_start, _date_end = self._build_date_string(_start_day, _end_day)
						self.add_event(_start_name, _date_start, _date_end)
						return

					_day = {'year': year, 'month': month, 'day': day}
					_name = self.data[_day['year']][_day['month']].get(_day['day'])
					if(not _start_day):
						_start_day = _end_day = _day
						_start_name = _name
					if(_start_name != _name):
						_date_start, _date_end = self._build_date_string(_start_day, _end_day)
						if (_start_name):
							self.add_event(_start_name, _date_start, _date_end)
						_start_day = _end_day = _day
						_start_name = _name
					else:
						_end_day = _day

	def is_last_day_of_month(self, day, days):
		return days.index(day) == (len(days) - 1)

	def _build_date_string(self, start_day, _end_day):
		'''
		stringify the date to be used in calendar event
		'''
		end_day_year, end_day_month, end_day_day = self.cal.plus_one_day(_end_day.get('year'), _end_day.get('month'), _end_day.get('day'))
		return "%s%02d%02d" % (start_day.get('year'), start_day.get('month'), start_day.get('day'),), "%s%02d%02d" % (end_day_year, end_day_month, end_day_day,)

	def add_event(self, name, start, end):
		'''
		add event to self.calendar
		'''
		_event = {'name': name, 'begin': start, 'end': end}
		if _event in self.event_list:
			# we don't like duplicates
			return
			
		self.event_list.append(_event)
		# event = Event(name=name, begin=start, end=end)		# not working atm
		# event.make_all_day()								# not working atm
		if not name:
			event = Event(name="24-7 NOT ASSIGNED")
		else:
			event = Event(name="24-7 %s" % (name,))
		event.properties = {'X-MICROSOFT-CDO-ALLDAYEVENT': 'TRUE', "DTSTART;VALUE=DATE": start, "DTEND;VALUE=DATE": end, "CLASS": "PUBLIC"}
		self.calendar.events.append(event)

	def save_to_file(self):
		with open(self.filename, 'w') as cal_file:
			cal_file.writelines(self.calendar)

	def generate_ical_file(self):
		self.build_cal_structure()
		self.save_to_file()

	def generate_and_export_to_ox(self):
		self.build_cal_structure()
		self.dav.parse_and_upload_data(self.calendar)


@Event._outputs
def o_properties(event, container):
	""" Workaround for arbitrary properties on Events.
	"""
	if hasattr(event, 'properties') and event.properties:
		for key in event.properties.iterkeys():
			container.append(ContentLine(key, value=event.properties.get(key)))


def read_config():
	config = ConfigParser.ConfigParser()
	config.readfp(open(CALDAV_CREDENTIALS_FILE))
	return config

if __name__ == '__main__':

	check_needed_files()
	config = read_config()
	cal = Cal()
	sheet = Spreadsheet()
	date_range = cal.get_calendar_month_range()
	spreadsheet_data = {}

	for year in date_range.iterkeys():
		data = sheet.get_data_for_months(year, date_range[year])
		spreadsheet_data[year] = data

	ical = Ical(spreadsheet_data, cal, config)
	ical.generate_and_export_to_ox()
	#ical.generate_ical_file()
