#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
#
#
#
# Dependencies:  gspread, oauth2client

import csv
import gspread
import json
import pgdb
import optparse
import os
import time
from calendar import TimeEncoding, month_name
from datetime import date, datetime, timedelta
from dateutil.easter import easter
from oauth2client.client import SignedJwtAssertionCredentials
try:
	from collections import OrderedDict as od
except ImportError:
	# python 2.6 or earlier, use backport
	from ordereddict import OrderedDict as od

# TODO all this should go to a config file or ucr in the future

DATABASE = 'univention'
HOST = 'openerp.knut.univention.de:5432'
USER = 'openerp'
SECRET_FILENAME = os.path.join(os.getenv('HOME'), '.openerp.secret')
ENCODING = 'de_DE.UTF-8'

FIRST_TICKET_MIN_TIME = 2   # minimal time the first assignment gets expanded to

WEEK_VALUE = 250            # base per week value if support
CHRISTMAS_ADDITION = 50     # if christmas or silvester week this is added per day
VALUE_6_22 = 50
VALUE_22_6 = 62.5
VALUE_6_22_HOLIDAY = 65     # if holiday
VALUE_22_6_HOLIDAY = 75     # if holiday

SUPPORT_22_6_ID = 7246
SUPPORT_22_6_NAME = '33586:Support-IN8-22_6'
SUPPORT_6_22_ID = 7247
SUPPORT_6_22_NAME = '33586:Support-IN8-18_22'

ACCOUNTS = [SUPPORT_22_6_NAME, SUPPORT_6_22_NAME]
ACCOUNTS_22_6 = [SUPPORT_22_6_NAME]
ACCOUNTS_6_22 = [SUPPORT_6_22_NAME]


GOOGLE_CREDENTIALS_FILE = os.path.join(os.getenv('HOME'), '.google.json')
GOOGLE_KEY = '1i0DOGO9AXg8kKq0W7xFLqwgse1tav1wgxtKkx6XCCqA' #NEW 
#GOOGLE_KEY = '11r_98CcYxf8zN7-D3KejM2mxhUyhft4TvJ9RAJlqnz4' #OLD
JSON_KEY = json.load(open(GOOGLE_CREDENTIALS_FILE))
SCOPE = ['https://spreadsheets.google.com/feeds']

SEND_MAIL = True
MAIL_RECIPIENTS = ['24-7-reporting@univention.de']
MAIL_FROM = '24-7-reporting@univention.de'
DEL_AFTER_SENDING = True

VERBOSE = None


class Calendar:

	def __init__(self, year, month):
		self.year = int(year)
		self.month = int(month)
		self.state = "HB"
		self.holidays = self.get_public_holiday(self.year)
		self.special_days = self.get_special_days(self.year)
		self.month_range = self.get_month_start_end(self.year, self.month)
		self.cal_weeks = self.divide_month_into_weeks(self.month_range)
		self.day_count

	def get_public_holiday(self, year):
		easter_day = easter(year)
		public_holiday = (
			date(year, 01, 01),                # Neujahr
			easter_day + timedelta(-2),        # Karfreitag
			easter_day + timedelta(1),         # Ostermontag
			date(year, 05, 01),                # Tag der Arbeit
			easter_day + timedelta(39),        # Christi Himmelfahrt
			easter_day + timedelta(50),        # Pfingstmontag
			date(year, 10, 03),                # Tag der Deutschen Einheit
			date(year, 12, 25),                # 1. Weihnachtstag
			date(year, 12, 26)                 # 2. Weihnachtstag
		)
		return public_holiday

	def get_special_days(self, year):
		special_days = (
			date(year, 12, 25),
			date(year, 12, 26),
			date(year, 12, 31)
		)
		return special_days

	def get_month_start_end(self, year, month):
		_year = int(year)
		_month = int(month)
		date_range = {}
		date_range['start'] = date(_year, _month, 1)
		# quick fix for turn of the year
		if _month == 12:
			date_range['end'] = date(_year + 1, 1, 1) + timedelta(-1)
		else:
			date_range['end'] = date(_year, _month + 1, 1) + timedelta(-1)
		return date_range

	def date_range(self, start_date, end_date):
		if isinstance(start_date, datetime):
			start_date = start_date.date()
		if isinstance(end_date, datetime):
			end_date = end_date.date()
		self.day_count = (end_date - start_date).days + 1
		for n in range(self.day_count if self.day_count else 1):
			yield start_date + timedelta(n)

	def divide_month_into_weeks(self, month_range):
		month_start = month_range['start']
		month_end = month_range['end']
		weeks = {}
		for day in self.date_range(month_start, month_end):
			week_nr = str(int(day.strftime("%W")) + 1)
			if week_nr not in weeks:
				weeks[week_nr] = {}
				weeks[week_nr]['days'] = []
				weeks[week_nr]['holidays'] = []
			if day in self.holidays:
				weeks[week_nr]['holidays'].append(day)
			else:
				weeks[week_nr]['days'].append(day)
		return weeks

	def get_month_name(self, month_no, locale=ENCODING):
		if isinstance(month_no, basestring):
			month_no = int(month_no)
		with TimeEncoding(locale) as encoding:
			name = month_name[month_no]
			if encoding is not None:
				name = name.decode(encoding)
			return name


class Connection:

	def __init__(self):
		self.database = DATABASE
		self.host = HOST
		self.user = USER
		self.password = self.get_openerp_password()

	def make_query(self, query_string):
		erp_connection = pgdb.connect(database=self.database, host=self.host, user=self.user, password=self.password)
		erp_cursor = erp_connection.cursor()
		erp_cursor.execute(query_string)
		erp_result = erp_cursor.fetchall()
		erp_connection.close()
		if erp_result:
			return erp_result
		else:
			return None

	def get_openerp_password(self):
		if not os.path.exists(SECRET_FILENAME):
			raise Exception('E: OpenERP secret file not found. Please copy the password of the openerp user to %s' % SECRET_FILENAME)
		return open(SECRET_FILENAME).read().strip(' \n')


class AZEData:

	def __init__(self):
		self.connection = Connection()
		self.raw_data = self.get_data_for_period()
		self.ordered_by_day = self.order_by_day()

	def get_data_for_period(self):
		interactive_print("Gathering AZE data")
		data = None
		period = cal.month_range
		start = period['start'].strftime("%d.%m.%Y")
		end = period['end'].strftime("%d.%m.%Y")

		query_string = """SELECT t2.name as Person, t1.name as Konto, t0.date as Tag, t0.hours as Stunden
		FROM project_task_work t0
		LEFT JOIN project_task t1 ON (t0.task_id = t1.id)
		LEFT JOIN res_users t2 ON (t0.user_id = t2.id)
		WHERE (t1.id = {0} OR t1.id = {1}) AND (t0.date >= '{2}' AND t0.date <= '{3}')
		GROUP BY Tag, Person, Konto, Stunden
		ORDER BY Tag""".format(SUPPORT_22_6_ID, SUPPORT_6_22_ID, start, end, )
		data = self.connection.make_query(query_string)
		return data

	def order_by_day(self):
		by_day = {}
		if self.raw_data:
			for entry in self.raw_data:
				agent = entry[0]
				raw_account = entry[1]
				if raw_account.startswith('Support'):
					account = raw_account.split(':')[1].lstrip()
				else:
					account = raw_account
				_date = entry[2].split(' ')[0]
				_day_values = _date.split('-')
				_day = date(int(_day_values[0]), int(_day_values[1]), int(_day_values[2]))
				day = _day.day
				time = entry[3]
				if day not in by_day:
					by_day[day] = {}
				if agent not in by_day[day]:
					by_day[day][agent] = {}
					if time < FIRST_TICKET_MIN_TIME:
						time = FIRST_TICKET_MIN_TIME
					by_day[day][agent][account] = time
				else:
					if account in by_day[day][agent].keys():
						booked_time = by_day[day][agent][account]
						time += booked_time
					by_day[day][agent][account] = time
		return by_day


class Spreadsheet:

	def __init__(self, year, month):
		self.year = year
		self.month = month
		self.credentials = SignedJwtAssertionCredentials(JSON_KEY['client_email'], JSON_KEY['private_key'].encode(), SCOPE)
		self.gc = gspread.authorize(self.credentials)
		self.sheet = self.gc.open_by_key(GOOGLE_KEY)
		self.worksheet = self.get_worksheet(self.year)
		self.named_month = cal.get_month_name(self.month)

	def gather_data(self):
		interactive_print("Gathering spreadsheet data")
		data = self.get_months_data(self.named_month)
		return data

	def get_worksheet(self, year):
		sheet = None
		try:
			sheet = self.sheet.worksheet(year)
		except:
			raise Exception("Could not get Spreadsheet Data")
		return sheet

	def get_months_data(self, named_month):
		base_cell = self.worksheet.find(named_month)
		data = {}
		days = cal.day_count
		for day in range(1, days + 1):
			data[day] = self.worksheet.cell(base_cell.row + day, base_cell.col).value
		return data


class Values:

	def build_week_values(self, week):
		week_values = {}
		base_day_value = float(WEEK_VALUE) / 7
		if week.get('holidays'):
			for holiday in week.get('holidays'):
				if holiday in cal.special_days:
					base_day_value = float(WEEK_VALUE) / 7 + CHRISTMAS_ADDITION
				week_values[holiday.day] = {}
				for account in ACCOUNTS_6_22:
					week_values[holiday.day][account] = VALUE_6_22_HOLIDAY
				for account in ACCOUNTS_22_6:
					week_values[holiday.day][account] = VALUE_22_6_HOLIDAY
		if week.get('days'):
			for day in week.get('days'):
				if day in cal.special_days:
					base_day_value = float(WEEK_VALUE) / 7 + CHRISTMAS_ADDITION
				week_values[day.day] = {}
				for account in ACCOUNTS_6_22:
					week_values[day.day][account] = VALUE_6_22
				for account in ACCOUNTS_22_6:
					week_values[day.day][account] = VALUE_22_6
		week_values['base_day_value'] = base_day_value

		return week_values


class Report:

	def __init__(self):
		self.spreadsheet_data = sheet.gather_data()
		self._aze_data = aze_data.ordered_by_day

		self.values = Values()
		interactive_print("Generating report")
		self.processed_data = self.join_data()
		self.csv_data = self.create_csv_data()

	def join_data(self):
		month_data = {}
		cal_weeks = cal.cal_weeks
		for week_nr in set(cal_weeks):
			week_values = self.values.build_week_values(cal_weeks[week_nr])
			week_days = cal_weeks[week_nr]['days'] + cal_weeks[week_nr]['holidays']

			for _day in week_days:
				day = int(_day.day)
				agent = self.spreadsheet_data.get(day)
				agents = [agent]

				_aze_values = self._aze_data.get(day)
				if _aze_values:

					# add agents who actually booked support
					for aze_agent in self._aze_data[day].keys():
						if aze_agent not in agents:
							agents.append(aze_agent)

				if not agents:
					continue

				for agent in agents:
					if agent not in month_data.keys():
						month_data[agent] = {}

					if week_nr not in month_data[agent].keys():
						month_data[agent][week_nr] = {}

					month_data[agent][week_nr][day] = {
						'base': week_values['base_day_value'],
					}

					if _aze_values:
						_aze_day_values = _aze_values.get(agent)
						if _aze_day_values:
							for account in _aze_day_values:
								time = _aze_day_values[account]
								month_data[agent][week_nr][day][account] = {}
								month_data[agent][week_nr][day][account]['time'] = time
								month_data[agent][week_nr][day][account]['value'] = float(time) * week_values[day][account]

		return month_data

	def create_csv_data(self):
		data = {}
		raw_data = od(sorted(self.processed_data.items()))
		for agent in raw_data:
			if not agent:
				continue
			_processed_agent_data = {}
			agent_data = od(sorted(raw_data[agent].items()))
			for week in agent_data:
				week_data = agent_data[week]
				_week = {
					'fixed_rate': 0,
					'euro': 0,
					'days': 0
				}
				for name in ACCOUNTS:
					_week[name] = 0

				for day in week_data:
					day_data = week_data[day]
					_week['days'] += 1
					_week['fixed_rate'] += day_data['base']
					for name in ACCOUNTS:
						if day_data.get(name):
							_week['euro'] += day_data[name]['value']
							_week[name] += day_data[name]['time']
				_week['euro'] += _week['fixed_rate']
				_week['fixed_rate'] = round(_week['fixed_rate'], 2)
				_week['euro'] = round(_week['euro'], 2)
				for name in ACCOUNTS:
					if not _week[name] == 0:
						_week[name] = round(_week[name], 2)

				_processed_agent_data[week] = _week
			_processed_agent_data = od(sorted(_processed_agent_data.items()))
			data[agent] = _processed_agent_data
		data = od(sorted(data.items()))
		return data

	def print_report(self, filename):
		with open(filename, 'wb') as csvfile:
			writer = csv.writer(csvfile, delimiter=';')
			data = self.csv_data
			if data:
				writer.writerow(("Name", "KW", "Tage", "Pauschale", "Zeit 6-22", "Zeit 22-6", "Euro"))
				for agent in data:
					for week in data[agent]:
						week_data = data[agent][week]
						if len(ACCOUNTS_6_22) == len(ACCOUNTS_22_6) == 1:
							row = [agent, week, week_data['days'], week_data['fixed_rate'], week_data[ACCOUNTS_6_22[0]], week_data[ACCOUNTS_22_6[0]], week_data['euro']]
						else:
							# this needs to be done
							pass
						writer.writerow(row)
		interactive_print("Finished writing report")

	def send_mail(self, filename, year, month):
		month_name = cal.get_month_name(month)
		for address in MAIL_RECIPIENTS:
			interactive_print("Sending mail to %s" % address)
			os.system('echo "Hallo, dies ist die 24-7-Support Zusammenfassung für %s %s" | mutt -e "my_hdr From: 24-7-reporting <%s>" -s "24-7-Support" -a %s -- %s' % (month_name, year, MAIL_FROM, filename, address))
		if DEL_AFTER_SENDING:
			interactive_print("All mails sent, deleting report")
			os.system("rm %s" % filename)


def check_needed_files():
	if not os.path.isfile(SECRET_FILENAME):
		raise Exception("%s not found, exiting." % (SECRET_FILENAME,))
	if not os.path.isfile(GOOGLE_CREDENTIALS_FILE):
		raise Exception("%s not found, exiting." % (GOOGLE_CREDENTIALS_FILE,))


def interactive_print(text):
	if VERBOSE:
		print(text)


if __name__ == '__main__':
	current_year = time.localtime().tm_year
	current_month = time.localtime().tm_mon

	check_needed_files()

	parser = optparse.OptionParser()
	parser.add_option("--year", dest="year", default=current_year, action="store", help="Reporting year")
	parser.add_option("--month", dest="month", default=None, action="store", help="Reporting month")
	parser.add_option("--filename", dest="filename", default=None, action="store", help="CSV filename, e.g. 'foo.csv'")
	parser.add_option("-v", "--verbose", dest="verbose", default=False, action="store_true", help="Display additional output and don't send mail")
	(options, args) = parser.parse_args()

	year = options.year
	month = None
	VERBOSE = options.verbose

	if options.month:
		month = options.month
	else:
		interactive_print("No month specified, defaulting to last month")
		if current_month > 1:
			month = current_month - 1
		else:
			year -= 1
			month = 12

	year = str(year)
	month = str(month)

	if options.filename:
		filename = options.filename
	else:
		filename = "24-7-reporting-%s-%s.csv" % (year, month)

	interactive_print("Starting reporting process for %s %s" % (year, month))

	cal = Calendar(year, month)

	aze_data = AZEData()

	sheet = Spreadsheet(year, month)

	rep = Report()
	rep. print_report(filename)

	if SEND_MAIL and not VERBOSE:
		rep.send_mail(filename, year, month)

	interactive_print("Report was written to %s" % filename)
